@@ -2,7 +2,7 @@ use strict;
 use warnings;
 package Device::CurrentCost;
 BEGIN {
-  $Device::CurrentCost::VERSION = '1.110390';
+  $Device::CurrentCost::VERSION = '1.110790';
 }
 
 # ABSTRACT: Perl modules for Current Cost energy monitors
@@ -24,15 +24,18 @@ use Time::HiRes;
 
 sub new {
   my ($pkg, %p) = @_;
-  croak $pkg.q{->new: 'device' parameter is required}
-    unless (exists $p{device});
   my $self = bless {
                     buf => '',
                     discard_timeout => 1,
                     type => CURRENT_COST_ENVY,
+                    history_callback => sub {},
                     %p
                    }, $pkg;
-  $self->open();
+  unless (exists $p{filehandle}) {
+    croak $pkg.q{->new: 'device' parameter is required}
+      unless (exists $p{device});
+    $self->open();
+  }
   $self;
 }
 
@@ -121,6 +124,7 @@ sub read {
     unless ($bytes) {
       croak((ref $self).'->read: '.(defined $bytes ? 'closed' : 'error: '.$!));
     }
+    print STDERR 'Read ', $bytes, "bytes\n" if DEBUG;
     $res = $self->read_one(\$self->{buf});
     return $res if (defined $res);
   } while (1);
@@ -130,17 +134,60 @@ sub read {
 sub read_one {
   my ($self, $rbuf) = @_;
   return unless ($$rbuf);
-  if ($$rbuf =~ s!(<msg>.*?</msg>)\s*!!) {
-    return Device::CurrentCost::Message->new(message => $1);
+  print STDERR 'Read one from !', $$rbuf, "!\n" if DEBUG;
+  if ($$rbuf =~ s!^\s*(<msg>.*?</msg>)\s*!!) {
+    my $msg = Device::CurrentCost::Message->new(message => $1);
+    my $t = $self->_time_now;
+    if ($msg->has_history) {
+      my $new = $msg->history;
+      my $our= $self->{history} || ($self->{history} = {});
+      foreach my $sensor (sort keys %$new) {
+        foreach my $interval (sort keys %{$new->{$sensor}}) {
+          foreach my $age (keys %{$new->{$sensor}->{$interval}}) {
+            $our->{$sensor}->{$interval}->{pending}->{$age} =
+              0+$new->{$sensor}->{$interval}->{$age};
+          }
+          if (exists $our->{$sensor}->{$interval}->{pending}->{1} ||
+              ($interval eq 'hours' &&
+               (exists $our->{$sensor}->{$interval}->{pending}->{4} ||
+                exists $our->{$sensor}->{$interval}->{pending}->{2}))) {
+            my $entries = keys %{$our->{$sensor}->{$interval}->{pending}};
+            if ($entries == { years => 4, months => 21, # envy
+                              days => 90, hours => 372 }->{$interval} ||
+                $entries == { years => 4, months => 12, # classic
+                              days => 31, hours => 13 }->{$interval}) {
+              %{$our->{$sensor}->{$interval}->{current}} =
+                %{$our->{$sensor}->{$interval}->{pending}};
+              $our->{$sensor}->{$interval}->{time} = $t;
+              $self->{history_callback}->($sensor, $interval,
+                   $our->{$sensor}->{$interval}->{current});
+            }
+            $our->{$sensor}->{$interval}->{pending} = {};
+          }
+        }
+      }
+    }
+    return $msg;
   } else {
     return;
   }
 }
 
+
+sub sensor_history {
+  my ($self, $sensor, $interval) = @_;
+  return unless (exists $self->{history}->{$sensor}->{$interval}->{current});
+  return {
+          time => $self->{history}->{$sensor}->{$interval}->{time},
+          data => $self->{history}->{$sensor}->{$interval}->{current}
+         };
+}
+
 sub _discard_buffer_check {
   my $self = shift;
   if ($self->{buf} ne '' &&
       $self->{_last_read} < ($self->_time_now - $self->{discard_timeout})) {
+    carp "Discarding '", $self->{buf}, "'";
     $self->{buf} = '';
   }
 }
@@ -160,7 +207,7 @@ Device::CurrentCost - Perl modules for Current Cost energy monitors
 
 =head1 VERSION
 
-version 1.110390
+version 1.110790
 
 =head1 SYNOPSIS
 
@@ -179,6 +226,14 @@ version 1.110390
                                          type => CURRENT_COST_CLASSIC);
   ...
 
+  open my $cclog, '<', 'currentcost.log' or die $!;
+  my $cc = Device::CurrentCost->new(filehandle => $cclog);
+
+  while (1) {
+    my $msg = $cc->read() or next;
+    print $msg->summary, "\n";
+  }
+
 =head1 DESCRIPTION
 
 Module for reading from Current Cost energy meters.
@@ -201,8 +256,14 @@ supported parameters are:
 
 =item device
 
-The name of the device to connect to.  The value should be a tty device
-name, e.g. C</dev/ttyUSB0>.  This parameter is mandatory.
+The name of the device to connect to.  The value should be a tty
+device name, e.g. C</dev/ttyUSB0> but a pipe or plain file should also
+work.  This parameter is mandatory if B<filehandle> is not given.
+
+=item filehandle
+
+A filehandle to read from.  This parameter is mandatory if B<device> is
+not given.
 
 =item type
 
@@ -214,6 +275,13 @@ C<CURRENT_COST_ENVY>.  The default is C<CURRENT_COST_ENVY>.
 The baud rate for the device.  The default is derived from the type and
 is either C<57600> (for Envy) or C<9600> (for classic).
 
+=item history_callback
+
+A function, taking a sensor id, a time interval and a hash reference
+of data as arguments, to be called every time a new complete set of
+history data becomes available.  The data hash reference has keys of
+the number of intervals ago and values of the reading at that time.
+
 =back
 
 =head2 C<device()>
@@ -257,6 +325,18 @@ insufficient data is available then undef is returned.
 
 B<IMPORTANT:> This API is still subject to change.
 
+=head2 C<sensor_history($sensor, $interval)>
+
+This method returns the most recent complete sensor data for the
+given sensor and the given interval (where interval must be one
+of 'hours', 'days', 'months' or 'years').  The return value is
+a hash reference with keys for 'time' and 'data'.  The 'time'
+value is the time (in seconds since epoch).  The 'data' value
+is a hash reference with keys of the number of intervals ago
+and values of the reading at that time.
+
+It returns undef if no history data has been received yet.
+
 =head1 AUTHOR
 
 Mark Hindess <soft-cpan@temporalanomaly.com>
